from __future__ import annotations
from typing import AsyncIterable, Union, \
  TYPE_CHECKING, cast
from inspect import iscoroutinefunction
from pathlib import Path
import io

from aiofile import AIOFile, LineReader

from .types import Final

if TYPE_CHECKING:  # keep mypy quiet
  from .path import AsyncPath


BEGINNING: Final[int] = 0
CHUNK_SIZE: Final[int] = 4 * 1_024

SEP: Final[str] = '\n'
ENCODING: Final[str] = 'utf-8'
ERRORS: Final[str] = 'replace'


Paths = Union['AsyncPath', Path, str]


class IterableAIOFile(AIOFile):
  def __init__(
    self,
    *args,
    errors: str | None = ERRORS,
    newline: str | None = SEP,
    **kwargs
  ):
    super().__init__(*args, **kwargs)
    self._errors: str | None = errors
    self._newline: str | None = newline

  def __aiter__(self) -> AsyncIterable[str]:
    encoding, errors, line_sep = self._get_options()

    return read_lines(
      self.name,
      line_sep,
      encoding=encoding,
      errors=errors,
    )

  def _get_options(
    self,
    encoding: str | None = None,
    errors: str | None = None
  ) -> tuple[str, str, str]:
    encoding = encoding or self.encoding or ENCODING
    errors = errors or self._errors or ERRORS
    line_sep: str = self._newline or SEP

    return encoding, errors, line_sep

  async def read_text(
    self,
    encoding: str | None = None,
    errors: str | None = None
  ) -> str:
    encoding, errors, line_sep = self._get_options(encoding, errors)

    return await read_full_file(
      self.name,
      line_sep,
      encoding=encoding,
      errors=errors
    )


async def read_lines(
  path: Paths,
  line_sep: str = SEP,
  chunk_size: int = CHUNK_SIZE,
  offset: int = BEGINNING,
  encoding: str = ENCODING,
  errors: str = ERRORS,
  **kwargs
) -> AsyncIterable[str]: 
  if hasattr(path, 'resolve'):
    if iscoroutinefunction(path.resolve):
      path = str(await path.resolve())

    else:
      path = str(path.resolve())

  path = cast(str, path)

  async with AIOFile(path, 'rb') as handle:
    reader = LineReader(
      handle,
      line_sep=line_sep,
      chunk_size=chunk_size,
      offset=offset
    )

    while line := await reader.readline():
      yield line.decode(encoding, errors=errors)


async def read_full_file(
  path: Paths,
  line_sep: str = SEP,
  chunk_size: int = CHUNK_SIZE,
  offset: int = BEGINNING,
  encoding: str = ENCODING,
  errors: str = ERRORS,
  **kwargs
) -> str:
  lines_gen = read_lines(
    path,
    line_sep=line_sep,
    chunk_size=chunk_size,
    offset=offset,
    encoding=encoding,
    errors=errors
  )

  with io.StringIO() as string:
    async for line in lines_gen:
      string.write(line)

    return string.getvalue()



from __future__ import annotations
from pathlib import _PosixFlavour, _WindowsFlavour, PurePath
from typing import Callable, Awaitable, TYPE_CHECKING
from errno import EINVAL
import os

from aiofiles.os import wrap as wrap_async

try:
  from pathlib import _getfinalpathname
  _getfinalpathname = wrap_async(_getfinalpathname)

except ImportError:
  async def _getfinalpathname(*args, **kwargs):
    raise ImportError("_getfinalpathname() requires a Windows/NT platform")

if TYPE_CHECKING:  # keep mypy quiet
  from .path import AsyncPath


getcwd: Callable[[], Awaitable[str]] = wrap_async(os.getcwd)


class _AsyncPosixFlavour(_PosixFlavour):
  async def gethomedir(self, username: str) -> str:
    gethomedir: Callable[[str], Awaitable[str]] = wrap_async(
      super().gethomedir
    )

    return await gethomedir(username)

  async def resolve(
    self,
    path: AsyncPath,
    strict: bool = False
  ) -> str | None:
    sep: str = self.sep
    accessor: '_AsyncAccessor' = path._accessor
    seen: dict[str, str | None] = {}

    async def _resolve(path: str, rest: str) -> str:
      if rest.startswith(sep):
        path = ''

      for name in rest.split(sep):
        if not name or name == '.':
          # current dir
          continue

        if name == '..':
            # parent dir
          path, _, _ = path.rpartition(sep)
          continue

        if path.endswith(sep):
          newpath = path + name

        else:
          newpath = path + sep + name

        if newpath in seen:
          # Already seen this path
          path = seen[newpath]
          if path is not None:
            # use cached value
            continue

          # The symlink is not resolved, so we must have a symlink loop.
          raise RuntimeError(f"Symlink loop from {newpath}")

        # Resolve the symbolic link
        try:
          target = await accessor.readlink(newpath)

        except OSError as e:
          if e.errno != EINVAL and strict:
            raise
          # Not a symlink, or non-strict mode. We just leave the path
          # untouched.
          path = newpath
        else:
          seen[newpath] = None # not resolved symlink
          path = await _resolve(path, target)
          seen[newpath] = path # resolved symlink

      return path
    # NOTE: according to POSIX, getcwd() cannot contain path components
    # which are symlinks.
    base = '' if path.is_absolute() else await getcwd()
    result = await _resolve(base, str(path))
    return result or sep


class _AsyncWindowsFlavour(_WindowsFlavour):
  async def gethomedir(self, username: str) -> str:
    gethomedir: Callable[[str], Awaitable[str]] = wrap_async(
      super().gethomedir
    )

    return await gethomedir(username)

  async def resolve(
    self,
    path: 'AsyncPath',
    strict: bool = False
  ) -> str | None:
    s = str(path)

    if not s:
      return await getcwd()

    previous_s: str | None = None

    if _getfinalpathname is not None:
      if strict:
        return self._ext_to_normal(await _getfinalpathname(s))
      else:
        tail_parts: List[str] = []  # End of the path after the first one not found
        while True:
          try:
            s = self._ext_to_normal(await _getfinalpathname(s))
          except FileNotFoundError:
            previous_s = s
            s, tail = os.path.split(s)
            tail_parts.append(tail)
            if previous_s == s:
              return path
          else:
            return os.path.join(s, *reversed(tail_parts))
    # Means fallback on absolute
    return None


_async_windows_flavour = _AsyncWindowsFlavour()
_async_posix_flavour = _AsyncPosixFlavour()



from typing import Callable, Any, Awaitable
from functools import wraps, partial
from asyncio import to_thread
import contextvars

from aiofiles.os import wrap as method_as_method_coro, \
  wrap as func_as_corofunc


CoroutineResult = Awaitable[Any]
CoroutineFunction = Callable[..., CoroutineResult]
CoroutineMethod = Callable[..., CoroutineResult]


def func_as_method_coro(func: Callable) -> CoroutineMethod:
  @wraps(func)
  async def method(self, *args, **kwargs) -> Any:
    return await to_thread(func, *args, **kwargs)

  return method


def coro_as_method_coro(coro: CoroutineFunction) -> CoroutineMethod:
  @wraps(coro)
  async def method(self, *args, **kwargs) -> Any:
    return await coro(*args, **kwargs)

  return method



# export class hierarchy
from .path import AsyncPath, AsyncPurePath, AsyncWindowsPath, \
  AsyncPosixPath, AsyncPurePosixPath, AsyncPureWindowsPath

# export iterable async file handle
from .handle import IterableAIOFile



__author__ = "Alex DeLorenzo"
from typing import List
from setuptools import setup
from pathlib import Path


PKG_NAME = "aiopath"
NAME = 'aiopath'
VERSION = "0.6.4"
LICENSE = "LGPL-3.0"

DESC = "📁 Async pathlib for Python"

REQUIREMENTS: List[str] = \
  Path('requirements.txt') \
    .read_text() \
    .splitlines()

README: str = Path('README.md').read_text()


setup(
  name=PKG_NAME,
  version=VERSION,
  description=DESC,
  long_description=README,
  long_description_content_type="text/markdown",
  url='https://github.com/alexdelorenzo/aiopath',
  author=__author__,
  license=LICENSE,
  packages=[NAME],
  zip_safe=True,
  install_requires=REQUIREMENTS,
  python_requires='>=3.10',
  include_package_data=True
)
